'''Fix zero diameters

Re-implementation of: https://bbpcode.epfl.ch/source/xref/sim/MUK/apps/Fix_Zero_Diameter.cpp
with sections recursion instead of point recursion
'''
from collections import namedtuple
import numpy as np

SMALL = 0.0001

Point = namedtuple('Point',
                   ['section',
                    'point_id'])  # index of the point within the given section


def _next_point_upstream(point):
    '''Yield upstream points until reaching the root'''
    section, point_id = point
    while not (section.is_root and point_id == 0):
        if point_id > 0:
            point_id -= 1
        else:
            section = section.parent
            point_id = len(section.points) - 1
        yield Point(section, point_id)


def _get_point_diameter(point):
    return point.section.diameters[point.point_id]


def _set_point_diameter(point, new_diameter):
    '''Set a given diameter

    Unfortunately morphio does not support single value assignment
    so one has to update the diameters for the whole sections
    '''
    diameters = point.section.diameters
    diameters[point.point_id] = new_diameter
    point.section.diameters = diameters


def _connect_average_legacy(from_point):
    '''Apply a ramp diameter between the two points

    Re-implementation of https://bbpcode.epfl.ch/source/xref/sim/MUK/muk/Zero_Diameter_Fixer.cpp#232
    '''
    count = 0
    next_point = None
    for next_point in _next_point_upstream(from_point):
        count += 1
        if _get_point_diameter(next_point) > SMALL:
            break
    if next_point is None or count <= 1:
        return

    from_diam = _get_point_diameter(from_point)
    to_diam = _get_point_diameter(next_point)
    step_diam = (to_diam - from_diam) / count
    for step_num, point in enumerate(_next_point_upstream(from_point), 1):
        if _get_point_diameter(point) <= SMALL:
            _set_point_diameter(point, from_diam + step_diam * step_num)


def _connect_average(from_point):
    '''Apply a ramp diameter between the two points

    Re-implementation of https://bbpcode.epfl.ch/source/xref/sim/MUK/muk/Zero_Diameter_Fixer.cpp#232
    Contrary to the previous implementation the diameter the ramp is computed in term of
    pathlength and no longer in term of point number
    '''
    prev_point = from_point
    pathlengths = []
    next_point = None
    for next_point in _next_point_upstream(from_point):
        pathlengths.append(np.linalg.norm(prev_point.section.points[prev_point.point_id] -
                                          next_point.section.points[next_point.point_id]))
        if _get_point_diameter(next_point) > SMALL:
            break
        prev_point = next_point
    if next_point is None or len(pathlengths) <= 1:
        return

    from_diam = _get_point_diameter(from_point)
    to_diam = _get_point_diameter(next_point)
    cumulative_pathlengths = np.cumsum(pathlengths)
    pathlength_fractions = cumulative_pathlengths / cumulative_pathlengths[-1]
    for point, fraction in zip(_next_point_upstream(from_point),
                               pathlength_fractions[:-1]):
        _set_point_diameter(point, from_diam + (to_diam - from_diam) * fraction)


def _fix_downstream(section):
    '''Re-implementation of recursePullFix() available at
    https://bbpcode.epfl.ch/source/xref/sim/MUK/muk/Zero_Diameter_Fixer.cpp#66

    If the current diameter is below the threshold, change its value to the biggest value
    among the 1-degree children downstream diameters

    This is recursive only until a diameter above threshold is found so this
    fixes zero diameters that are located at the beginning of the neurite

    Fixes this:
    Soma--0--0--0--1--1--1
    But not this:
    Soma--1--1--1--0--0--0--1--1

    Args:
        section(morphio.Section):
    '''
    nonzero_indices = np.asarray(section.diameters > SMALL).nonzero()[0]
    if len(nonzero_indices) > 0:
        idx = nonzero_indices[0]
        diameter = section.diameters[idx]
        section.diameters = np.concatenate((np.repeat(diameter, idx), section.diameters[idx:]))
        return diameter
    else:
        child_diameters = [_fix_downstream(child) for child in section.children]
        max_child_diameter = max(child_diameters) if child_diameters else 0
        if max_child_diameter > SMALL:
            section.diameters = np.repeat(max_child_diameter, len(section.diameters))
        return max_child_diameter


def _fix_in_between(section, stack, legacy):
    '''Re-implementation of
    https://bbpcode.epfl.ch/source/xref/sim/MUK/muk/Zero_Diameter_Fixer.cpp#162

    Fix diameters between two points with valid diameters by applying a ramp

    Args:
        point: the current points
        stack: the stack of upstream points
        legacy: whether to use legacy algorithm that didn't account for path lengths
    '''
    _connect = _connect_average
    if legacy:
        _connect = _connect_average_legacy

    nonzero_indices = np.asarray(section.diameters > SMALL).nonzero()[0]
    if len(nonzero_indices) > 0:
        if len(stack) > 0:
            _connect(Point(section, nonzero_indices[0]))
        # account for zero diameters inside `section`
        for i in range(1, len(nonzero_indices)):
            if nonzero_indices[i] - nonzero_indices[i - 1] > 1:
                _connect(Point(section, nonzero_indices[i]))
        stack.append(Point(section, nonzero_indices[-1]))
    for child in section.children:
        _fix_in_between(child, stack, legacy)
    if len(nonzero_indices) > 0:
        stack.pop()


def _fix_upstream(section, upstream_good_diameter):
    '''Re-implementation of recursePushFix() available at
    https://bbpcode.epfl.ch/source/xref/sim/MUK/muk/Zero_Diameter_Fixer.cpp#94

    Reset the diameter to upstream_good_diameter if the current value and all child values
    are below threshold

    Args:
        section: the current section
        upstream_good_diameter: the diameter value coming from upstream that will be used if
            the current diameter is not suitable

    Returns:
        The current diameter if above threshold, else the smallest child value above threshold
        else (if no child value is above threshold) returns 0
    '''
    diameters = section.diameters
    nonzero_indices = np.asarray(diameters > SMALL).nonzero()[0]
    last_nonzero_idx = nonzero_indices[-1] if len(nonzero_indices) > 0 else -1
    if last_nonzero_idx >= 0:
        upstream_good_diameter = diameters[last_nonzero_idx]
    child_diameters = [_fix_upstream(child, upstream_good_diameter) for child in section.children]
    max_child_diameter = max(child_diameters) if child_diameters else 0
    if max_child_diameter < SMALL:
        section.diameters = np.concatenate((
            diameters[:last_nonzero_idx + 1],
            np.repeat(upstream_good_diameter, len(diameters) - (last_nonzero_idx + 1))
        ))
    return max_child_diameter


def fix_neurite(root_section, legacy=False):
    '''Apply all fixes to a neurite'''
    _fix_downstream(root_section)
    _fix_in_between(root_section, list(), legacy)
    _fix_upstream(root_section, 0)


def fix_zero_diameters(neuron, legacy=False):
    '''Fix zero diameters'''
    for root in neuron.root_sections:
        fix_neurite(root, legacy)



from pathlib import Path

import numpy as np
from morphio.mut import Morphology
from numpy.testing import assert_array_equal, assert_array_almost_equal
from morph_tool import diff

from neuror.zero_diameter_fixer import (_fix_downstream, _fix_upstream,
                                        _fix_in_between, fix_zero_diameters)

DATA_PATH = Path(__file__).parent / 'data'


def test_fix_in_between():
    neuron = Morphology(DATA_PATH / 'zero-diameter-middle-of-neurite.asc')
    root = neuron.root_sections[0]

    _fix_in_between(root, list(), False)
    assert_array_equal(root.diameters, np.array([2., 2., 2.2, 2.4, 2.8, 3.], dtype=np.float32))
    assert_array_equal(root.children[0].diameters, np.array([3, 3, 3, 2.8], dtype=np.float32))
    assert_array_equal(root.children[0].children[0].diameters,
                       np.array([2.8, 2.6, 2.4, 2.2, 2.], dtype=np.float32))


def test_fix_in_between_legacy():
    neuron = Morphology(DATA_PATH / 'zero-diameter-middle-of-neurite.asc')
    root = neuron.root_sections[0]

    _fix_in_between(root, list(), True)
    assert_array_equal(root.diameters,
                       np.array([2., 2., 2.25, 2.5, 2.75, 3.], dtype=np.float32))
    assert_array_almost_equal(root.children[0].diameters,
                              np.array([3., 3., 3., 2.833333], dtype=np.float32))
    assert_array_almost_equal(root.children[0].children[0].diameters,
                              np.array([2.666667, 2.5, 2.333333, 2.166667, 2.], dtype=np.float32))


def test_fix_from_downstream():
    neuron = Morphology(DATA_PATH / 'zero-diameter-middle-of-neurite.asc')
    root = neuron.root_sections[0]

    _fix_downstream(root)
    assert_array_equal(neuron.section(0).diameters,
                       np.array([2., 2., 1.e-6, 1.e-6, 1.e-6, 3.], dtype=np.float32),
                       "The diameters should not have been modified")

    neuron = Morphology(DATA_PATH / 'zero-diameter-beginning-of-neurite.swc')
    root = neuron.root_sections[0]

    assert_array_equal(root.diameters, [0., 0.])
    _fix_downstream(root)
    assert_array_equal(root.diameters, [0., 0.],
                       'Diameters should not have been changed '
                       'since the whole neurite has a null diameter')

    neuron = Morphology(DATA_PATH / 'zero-diameter-beginning-of-neurite.swc')
    root = neuron.root_sections[1]

    assert_array_equal(root.diameters, [0, 0, 4])
    _fix_downstream(root)
    assert_array_equal(root.diameters, [4, 4, 4])


def test_fix_from_upstream():
    neuron = Morphology(DATA_PATH / 'zero-diameter-end-of-neurite.swc')
    root = neuron.root_sections[0]
    leaf = root.children[0]

    _fix_upstream(root, 12)
    assert_array_equal(leaf.diameters, np.array([2, 2, 2, 2, 2], dtype=np.float32))


def test_fix_zero_diameters():
    neuron = Morphology(DATA_PATH / 'zero-diameter-end-of-neurite.swc')
    fix_zero_diameters(neuron)
    leaf = neuron.root_sections[0].children[0]
    assert_array_equal(leaf.diameters, np.array([2, 2, 2, 2, 2], dtype=np.float32))


def test_functional():
    functional_neuron = Morphology(DATA_PATH / 'compare-zero-diameter/original.h5')
    fix_zero_diameters(functional_neuron, legacy=True)
    expected = Morphology(DATA_PATH / 'compare-zero-diameter/cpp-fixed.h5')
    assert not diff(functional_neuron, expected)



'''Generate output plots'''
import logging
import os
from datetime import datetime

import matplotlib
import matplotlib.pyplot as plt  # noqa, pylint: disable=ungrouped-imports,wrong-import-order,wrong-import-position
import numpy as np
from matplotlib.backends.backend_pdf import (  # noqa, pylint: disable=ungrouped-imports,wrong-import-order,wrong-import-position
    PdfPages,
)
from neurom import geom, load_neuron
from neurom.view.view import plot_neuron

L = logging.getLogger('neuror')

try:
    from plotly_helper.neuron_viewer import NeuronBuilder
except ImportError as e:
    raise ImportError(
        'neuror[plotly] is not installed.'
        ' Please install it by doing: pip install neuror[plotly]') from e

matplotlib.use('Agg')


def get_common_bounding_box(neurons):
    '''Returns the bounding box that wraps all neurons'''
    common_bbox = geom.bounding_box(neurons[0])
    for neuron in neurons[1:]:
        bbox = geom.bounding_box(neuron)
        common_bbox[0] = np.min(np.vstack([common_bbox[0], bbox[0]]), axis=0)
        common_bbox[1] = np.max(np.vstack([common_bbox[1], bbox[1]]), axis=0)

    return common_bbox


def plot(neuron, bbox, subplot, title, **kwargs):
    '''2D neuron plot'''
    ax = plt.subplot(subplot, facecolor='w', aspect='equal')
    xlim = (bbox[0][0], bbox[1][0])
    ylim = (bbox[0][2], bbox[1][2])

    plot_neuron(ax, neuron, **kwargs)
    ax.set_title(title)
    ax.set_aspect('equal', adjustable='box')
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)


def _neuron_subplot(folders, f, pp, subplot, titles):
    kwargs = {'plane': 'xz'}
    fig = plt.figure()
    neurons = [load_neuron(os.path.join(folder, f)) for folder in folders]

    common_bbox = get_common_bounding_box(neurons)

    for i, (neuron, title) in enumerate(zip(neurons, titles)):
        plot(neuron, common_bbox, subplot + 1 + i, title, **kwargs)
    fig.suptitle(f)
    pp.savefig()


def view_all(folders, titles, output_pdf=None):
    '''Generate PDF report'''
    if not output_pdf:
        path = './plots'
        output_pdf = os.path.join(path, datetime.now().strftime("%Y_%m_%d_%H_%M_%S") + '.pdf')
        if not os.path.exists(path):
            os.mkdir(path)

    pp = PdfPages(output_pdf)

    files = os.listdir(folders[0])

    subplot = 100 + 10 * len(folders)
    for f in files:
        L.info(f)

        try:
            _neuron_subplot(folders, f, pp, subplot, titles)
        except Exception as e:  # pylint: disable=broad-except
            L.info("e: %s", e)
            L.info('failu: %s', f)
    pp.close()
    L.info('Done writing %s', output_pdf)


def plot_repaired_neuron(neuron, cut_points, plot_file=None):
    ''' Draw a neuron using plotly

    Repaired section are displayed with a different colors'''

    for mode in ['3d', 'xz']:
        builder = NeuronBuilder(neuron, mode, neuron.name, False)
        for section, offset in cut_points.items():
            builder.color_section(section, 'green', recursive=True, start_point=offset)

        if plot_file is not None:
            root, ext = os.path.splitext(plot_file)
            plot_file = '{}_{}{}'.format(root, mode, ext)

        builder.plot(show_link=False, auto_open=False, filename=plot_file)



from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
import pandas as pd
from morph_tool import diff
from neurom import load_neuron
from nose.tools import assert_false
from numpy.testing import assert_array_almost_equal, assert_array_equal

import neuror.unravel as test_module
from neuror.cut_plane.detection import CutPlane

DATA = Path(__file__).parent / 'data'

SIMPLE = load_neuron(DATA / 'simple.swc')

def test_get_principal_direction():
    direction = test_module._get_principal_direction([[0.,0,0], [1,1,2]])

    # Trick since the PCA can return a vector in either of the 2 orientations
    if direction[0] < 0:
        direction *= -1
    assert_array_almost_equal(direction, np.array([0.408248, 0.408248, 0.816497]))

    assert_array_almost_equal(test_module._get_principal_direction([[0., 0, 0],
                                                                    [10, -1, 0],
                                                                    [10, 1, 0]]),
                              np.array([1, 0, 0]))
def test_unravel():
    neuron, mapping = test_module.unravel(DATA / 'simple.asc')
    assert_array_almost_equal(neuron.root_sections[0].points,
                              np.array([[ 0.      ,  0.      ,  0.      ],
                                        [1.      , 1.      , 0.      ],
                                        [2.      , 2.      , 0.      ],
                                        [2.707107, 1.292893, 0.      ]]))

    assert_array_almost_equal(neuron.root_sections[0].children[0].points[0],
                              np.array([2.707107, 1.292893, 0.    ]))

    assert_array_almost_equal(mapping[['x0', 'y0', 'z0']].values,
                              [[0., 0., 0.],
                               [1., 1., 0.],
                               [2., 0., 0.],
                               [3., 0., 0.],
                               [3., 0., 0.],
                               [4., 1., 0.],
                               [3., 0., 0.],
                               [6., 4., 2.]])
    assert_array_almost_equal(mapping[['x1', 'y1', 'z1']].values,
                              [[0.        , 0.        , 0.        ],
                               [0.99999994, 0.99999994, 0.        ],
                               [1.99999988, 1.99999988, 0.        ],
                               [2.70710659, 1.29289317, 0.        ],
                               [2.70710659, 1.29289317, 0.        ],
                               [3.70710659, 2.29289317, 0.        ],
                               [2.70710659, 1.29289317, 0.        ],
                               [5.70710659, 5.29289341, 2.        ]])


def test_unravel_no_path_length():
    neuron, mapping = test_module.unravel(DATA / 'simple.asc', use_path_length=False)
    assert_array_almost_equal(neuron.root_sections[0].points,
                              np.array([[ 0.      ,  0.      ,  0.      ],
                                        [ 1.404784, -0.163042,  0.      ],
                                        [ 2.809567, -0.326085,  0.      ],
                                        [ 3.8029  , -0.441373,  0.      ]]))

    assert_array_almost_equal(neuron.root_sections[0].children[0].points[0],
                              np.array([ 3.8029  , -0.441373,  0.      ]))

    assert_array_almost_equal(mapping[['x0', 'y0', 'z0']].values,
                              [[0., 0., 0.],
                               [1., 1., 0.],
                               [2., 0., 0.],
                               [3., 0., 0.],
                               [3., 0., 0.],
                               [4., 1., 0.],
                               [3., 0., 0.],
                               [6., 4., 2.]])

    assert_array_almost_equal(mapping[['x1', 'y1', 'z1']].values,
                              [[0.        ,  0.        , 0.],
                               [1.40478373, -0.16304241, 0.],
                               [2.80956745, -0.32608482, 0.],
                               [3.8028996 , -0.44137323, 0.],
                               [3.8028996 , -0.44137323, 0.],
                               [4.80289936,  0.55862677, 0.],
                               [3.8028996 , -0.44137323, 0.],
                               [6.80289936,  3.55862665, 2.]])


def test_unravel_with_backward_segment():
    '''Test the fix to the  unravel issue
    which was not working when there was a segment going backward wrt to the window direction
    (direction from the window first to the window last point)
    '''
    neuron, mapping = test_module.unravel(DATA / 'simple-with-backward-segment.asc')
    assert_array_almost_equal(neuron.root_sections[0].points,
                              np.array([[0.        , 0.        , 0.        ],
                                        [0.99999994, 0.99999994, 0.        ],
                                        [1.9999999 , 1.9999999 , 0.        ],
                                        [3.4142134 , 3.4142134 , 0.        ],
                                        [4.828427  , 4.828427  , 0.        ],
                                        [5.5355334 , 4.1213202 , 0.        ]], dtype=np.float32))


def test_unravel_plane():
    with TemporaryDirectory('test-unravel-plane'):
        mapping = pd.read_csv(DATA / 'mapping.csv')
        input_plane = CutPlane.from_json(DATA / 'neuron-slice-plane.json')
        plane = test_module.unravel_plane(CutPlane.from_json(DATA / 'neuron-slice-plane.json'), mapping)
        assert_array_almost_equal(plane.cut_leaves_coordinates,
                                  [[-111.24885559,   -1.29032707,   55.46524429],
                                   [-156.59031677,   23.12454224,   53.51153946],
                                   [ -60.68390274,  111.23972321,   54.0868721 ],
                                   [  15.63226223,  -25.58386421,   51.45604706],
                                   [ -14.81918049,  -10.20301151,   52.09701157],
                                   [ -32.34645462,  -19.15251732,   50.921875  ],
                                   [ -40.98423767,  -20.00574303,   51.760952  ],
                                   [ -26.68880844,   -1.51575243,   51.23664093],
                                   [ -18.93479538,   22.15104103,   50.69124985],
                                   [  -7.5544219 ,   15.30504322,   51.06955719],
                                   [  32.2554512 ,   56.86440277,   49.46508026],
                                   [  -4.24387503,   47.21520996,   52.44573212]])

        input_plane.cut_leaves_coordinates = []
        assert_array_almost_equal(test_module.unravel_plane(input_plane, mapping),
                                  [])

        input_plane.cut_leaves_coordinates = None
        assert_array_almost_equal(test_module.unravel_plane(input_plane, mapping),
                                  [])

def test_unravel_all():
    with TemporaryDirectory('test-unravel-plane') as output:
        output = Path(output)

        input = DATA / 'input-unravel-all'
        raw_planes = input / 'raw-planes'
        unravel_planes = output / 'unravelled-planes'
        unravel_planes.mkdir()

        test_module.unravel_all(input, output, raw_planes, unravel_planes)
        assert_array_equal(list(output.rglob('*.h5')), [output / 'Neuron_slice.h5'])

def test_legacy():
    actual, _ = test_module.unravel(DATA / 'legacy-unravel/1-pt-soma.swc', legacy_behavior=True)
    assert_false(diff(actual, DATA / 'legacy-unravel/expected-1-pt-soma.h5'))

    actual, _ = test_module.unravel(DATA / 'legacy-unravel/3-pts-soma.swc', legacy_behavior=True)
    assert_false(diff(actual, DATA / 'legacy-unravel/expected-3-pts-soma.h5'))



"""neuror setup.py"""
import imp
import sys

from setuptools import setup, find_packages

if sys.version_info < (3, 6):
    sys.exit("Sorry, Python < 3.6 is not supported")

# read the contents of the README file
with open("README.rst", encoding='utf-8') as f:
    README = f.read()

setup(
    name='NeuroR',
    author='Blue Brain Project, EPFL',
    description='A morphology repair tool',
    long_description=README,
    long_description_content_type="text/x-rst",
    url="https://github.com/bluebrain/neuror",
    license="LGPLv3",
    install_requires=[
        'click>=0.7.0',
        'matplotlib>=2.2.3',
        'morph-tool>=2.4.4',
        'morphio>=3.0.0',
        'neurom>=2.0.0',
        'numpy>=1.19.2',
        'nptyping>=1.3.0',
        'pandas>=0.24.2',
        'pyquaternion>=0.9.2',
        'scipy>=1.2.0',
    ],
    extras_require={
        'plotly': [
            'dash-core-components>=0.46.0',  # HTML components
            'dash-table>=3.6.0',  # Interactive DataTable component (new!)
            'dash>=0.41.0',  # The core dash backend
            'plotly-helper>=0.0.2',
        ],
        'docs': [
            'sphinx-autorun>=1.1.1',
            'sphinx-bluebrain-theme>=0.2.4',
            'sphinx-click>=2.5.0',
            'sphinx>=2.0.0',
        ],
    },
    packages=find_packages(),
    entry_points={'console_scripts': ['neuror=neuror.cli:cli']},
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
)



'''The unravel module

Unravelling is the action of "stretching" the cell that
has been shrunk because of the dehydratation caused by the slicing'''
import json
import logging
import os
from pathlib import Path

import morphio
import numpy as np
import pandas as pd
from morph_tool.utils import iter_morphology_files
from neurom.morphmath import interval_lengths
from scipy.spatial.ckdtree import cKDTree

from neuror.cut_plane import CutPlane
from neuror.utils import RepairJSON

L = logging.getLogger('neuror')

DEFAULT_WINDOW_HALF_LENGTH = 5
DEFAULT_WINDOW_HALF_LENGTH_PATH = 8


def _get_principal_direction(points):
    '''Return the principal direction of a point cloud
    It is the eigen vector of the covariance matrix with the highest eigen value'''

    X = np.copy(np.asarray(points))
    X -= np.mean(X, axis=0)
    C = np.dot(X.T, X)
    w, v = np.linalg.eigh(C)
    return v[:, w.argmax()]


def _unravel_section(section, window_half_length, soma, legacy_behavior):
    '''Unravel a section using number of adjacent points as window_half_length'''
    # pylint: disable=too-many-locals
    points = section.points
    if legacy_behavior and section.is_root and len(soma.points) > 1:
        points = np.vstack((soma.points[0], points))
    point_count = len(points)
    if section.is_root:
        if legacy_behavior and len(soma.points) > 1:
            unravelled_points = [soma.points[0]]
        else:
            unravelled_points = [section.points[0]]
    else:
        unravelled_points = [section.parent.points[-1]]

    for window_center in range(1, point_count):
        window_start = int(max(0, window_center - window_half_length - 1))
        window_end = int(min(point_count, window_center + window_half_length + 1))
        direction = _get_principal_direction(points[window_start:window_end])

        segment = points[window_center] - points[window_center - 1]
        window_direction = points[window_end - 1] - points[window_start]

        # make it span length the same as the original window_direction within the window
        direction *= np.linalg.norm(segment)

        scalar_product = np.dot(window_direction, direction)
        # point it in the same direction as the window
        direction *= np.sign(scalar_product or 1.)

        point = direction + unravelled_points[-1]
        unravelled_points.append(point)

    section.points = unravelled_points
    if legacy_behavior and section.is_root and len(soma.points) > 1:
        section.diameters = np.hstack((soma.diameters[0], section.diameters))


# pylint: disable=too-many-locals
def _unravel_section_path_length(sec, window_half_length, soma, legacy_behavior):
    '''Unravel a section using path length as window_half_length'''
    unraveled_points = sec.points
    n_points = len(sec.points)

    if legacy_behavior and sec.is_root and len(soma.points) > 1:
        unraveled_points = np.vstack((soma.points[0], unraveled_points))

    # ensure the first point is the parent's last point
    if not sec.is_root:
        unraveled_points[0] = sec.parent.points[-1]

    for window_center in range(1, n_points):
        # find how many segement on the left and right to be close to window_half_length
        intervals = interval_lengths(sec.points)
        _l_left = np.cumsum(intervals[window_center::-1])
        _l_right = np.cumsum(intervals[min(n_points - 2, window_center):])
        window_start = np.argmin(abs(_l_left - window_half_length))
        window_end = np.argmin(abs(_l_right - window_half_length))

        # if we are near the first point of the section we increase window from the right/left
        # the 0.9 is only to not do that to often
        _left_length = _l_left[window_start]
        if _left_length < 0.9 * window_half_length:
            window_end = np.argmin(abs(_l_right - (2 * window_half_length - _left_length)))
        _right_length = _l_right[window_end]
        if _right_length < 0.9 * window_half_length:
            window_start = np.argmin(abs(_l_left - (2 * window_half_length - _right_length)))

        # center bounds to windows center
        window_start = window_center - window_start
        window_end = window_center + window_end

        # if windows as 0 width, extend by one segment on each side if possible
        if window_start == window_end:
            window_start = max(0, window_start - 1)
            window_end = min(n_points, window_end + 1)

        direction = _get_principal_direction(sec.points[window_start:window_end])
        original_segment = sec.points[window_center] - sec.points[window_center - 1]

        # make it span length the same as the original segment within the window
        direction *= np.linalg.norm(original_segment)

        # point it in the same direction as the window
        window_direction = sec.points[window_end - 1] - sec.points[window_start]
        scalar_product = np.dot(window_direction, direction)
        direction *= np.sign(scalar_product or 1.)

        # update the unravel points
        unraveled_points[window_center] = unraveled_points[window_center - 1] + direction

    sec.points = unraveled_points
    if legacy_behavior and sec.is_root and len(soma.points) > 1:
        sec.diameters = np.hstack((soma.diameters[0], sec.diameters))


def unravel(filename, window_half_length=None,
            legacy_behavior=False, use_path_length=True):
    '''Return an unravelled neuron

    Segment are unravelled iteratively
    Each segment direction is replaced by the averaged direction in a sliding window
    around this segment, preserving the original segment length.
    The start position of the new segment is the end of the latest unravelled segment

    Based initially on:
    DOI: 10.7551/mitpress/9780262013277.001.0001
    Section: 9.2 Repair of Neuronal Dendrites

    Args:
        filename (str): the neuron to unravel
        window_half_length (int): path length that defines half of the sliding window
        legacy_behavior (bool): if yes, when the soma has more than one point, the first point of
            the soma is appended to the start of each neurite.
        use_path_length (bool): if False, the argument window_half_length will be recasted to an int
            and correspond to number of points on each side of the window.

    Returns:
        a tuple (morphio.mut.Morphology, dict) where first item is the unravelled
            morphology and the second one is the mapping of each point coordinate
            before and after unravelling
    '''
    morph = morphio.Morphology(filename, options=morphio.Option.nrn_order)
    new_morph = morphio.mut.Morphology(morph, options=morphio.Option.nrn_order)  # noqa, pylint: disable=no-member

    if window_half_length is None:
        window_half_length = DEFAULT_WINDOW_HALF_LENGTH_PATH \
            if use_path_length else DEFAULT_WINDOW_HALF_LENGTH

    coord_before = np.empty([0, 3])
    coord_after = np.empty([0, 3])
    for sec, new_section in zip(morph.iter(), new_morph.iter()):
        if use_path_length:
            _unravel_section_path_length(
                new_section, window_half_length, morph.soma, legacy_behavior
            )
        else:
            _unravel_section(new_section, int(window_half_length), morph.soma, legacy_behavior)

        coord_before = np.append(coord_before, sec.points, axis=0)
        coord_after = np.append(coord_after, new_section.points, axis=0)

        if legacy_behavior and len(morph.soma.points) > 1:  # pylint: disable=no-member
            coord_before = np.vstack((morph.soma.points[0], coord_before))  # noqa, pylint: disable=no-member
        mapping = pd.DataFrame({
            'x0': coord_before[:, 0],
            'y0': coord_before[:, 1],
            'z0': coord_before[:, 2],
            'x1': coord_after[:, 0],
            'y1': coord_after[:, 1],
            'z1': coord_after[:, 2],
        })

    L.debug('Unravel successful for file: %s', filename)
    return new_morph, mapping


def unravel_plane(plane, mapping):
    '''Return a new CutPlane object where the cut-leaves
    position has been updated after unravelling'''
    leaves = plane.cut_leaves_coordinates

    if not np.any(leaves):
        return plane

    t = cKDTree(mapping[['x0', 'y0', 'z0']])
    distances, indices = t.query(leaves)
    not_matching_leaves = np.where(distances > 1e-5)[0]
    if not_matching_leaves.size:
        raise Exception('Cannot find the following leaves in the mapping:\n{}'.format(
                        leaves[not_matching_leaves]))
    plane.cut_leaves_coordinates = mapping.iloc[indices][['x1', 'y1', 'z1']].values
    return plane


def unravel_all(raw_dir, unravelled_dir,
                raw_planes_dir,
                unravelled_planes_dir,
                window_half_length=None):
    '''Repair all morphologies in input folder
    '''
    if not os.path.exists(raw_planes_dir):
        raise Exception('{} does not exist'.format(raw_planes_dir))

    if not os.path.exists(unravelled_planes_dir):
        os.mkdir(unravelled_planes_dir)

    for inputfilename in iter_morphology_files(raw_dir):
        L.debug('Unravelling: %s', inputfilename)
        outfilename = Path(unravelled_dir, inputfilename.name)
        raw_plane = CutPlane.from_json(
            Path(raw_planes_dir, inputfilename.name).with_suffix('.json'))
        unravelled_plane = Path(unravelled_planes_dir, inputfilename.name).with_suffix('.json')

        try:
            neuron, mapping = unravel(str(inputfilename), window_half_length)
            neuron.write(str(outfilename))
            with open(str(unravelled_plane), 'w') as f:
                json.dump(unravel_plane(raw_plane, mapping).to_json(), f, cls=RepairJSON)

        except Exception as e:  # noqa, pylint: disable=broad-except
            L.warning('Unravelling %s failed', f)
            L.warning(e, exc_info=True)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

from pkg_resources import get_distribution

# -- Project information -----------------------------------------------------

project = 'NeuroR'
version = get_distribution('neuror').version
release = version

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx_click.ext',
    'sphinx_autorun',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
]

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True
suppress_warnings = ["ref.python"]
autosummary_generate = True
autosummary_imported_members = True
autodoc_default_options = {
    'members': True,
    'imported-members': True,
    'show-inheritance': True,
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx-bluebrain-theme'
html_title = 'NeuroR'
html_show_sourcelink = False
html_theme_options = {
    "repo_url": "https://github.com/BlueBrain/NeuroR/",
    "repo_name": "BlueBrain/NeuroR"
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']



# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False
from sphinx.ext.autosummary import Autosummary

class AutosummaryOverride(Autosummary):
    """Extends Autosummary to ensure the nosignatures option is set."""

    def run(self):
        """Wrap the autodoc output in a div with autodoc class."""
        self.options["nosignatures"] = self.options.get("nosignatures", True)
        result = super(AutosummaryOverride, self).run()
        return result

def add_autosummary_override(app):
    """Override the autosummary definition to ensure no signatures."""
    if "sphinx.ext.autosummary" in app.extensions:
        app.add_directive("autosummary", AutosummaryOverride, override=True)

def allow_only_neuror(app, what, name, obj, skip, options):
    """Check that the member is part of neuror, exlude otherwise."""
    if what in {"module", "class", "exception", "function"} and "neuror" not in getattr(obj, "__module__", ""):
        return True
    return skip

def setup(app):
    app.connect('builder-inited', add_autosummary_override)
    app.connect('autodoc-skip-member', allow_only_neuror)



import json
from pathlib import Path

import neurom as nm
import numpy as np
from mock import MagicMock
from nose.tools import assert_almost_equal, assert_equal, assert_not_equal, assert_raises, ok_
from numpy.testing import assert_array_almost_equal, assert_array_equal

import neuror.cut_plane.detection as test_module
from neuror.cut_plane.detection import _minimize, _success_function

DATA = Path(__file__).parent.parent / 'data'


def _get_points():
    '''Utility function to get all neuron points'''
    neuron = nm.load_neuron(DATA / 'rotated.h5')
    return np.array([point
                     for neurite in neuron.neurites
                     for section in nm.iter_sections(neurite)
                     for point in section.points])


def test_project_normal():
    plane = test_module.PlaneEquation(0, 0, 1, 0)
    assert_array_equal(plane.project_on_normal(np.array([[0,0,2], [1,2,3]])),
                       [2, 3])

    plane = test_module.PlaneEquation(0, 1, 0, -9)
    assert_array_equal(plane.project_on_normal(np.array([[0,0,0], [1,2,3]])),
                       [-9, -7])


def test_projection():
    points = _get_points()
    plane = test_module.PlaneEquation.from_rotations_translations([4, 45, -21, 0, 0, 61])
    bin_width = 10
    n_left, n_right = plane.count_near_plane(points, bin_width)

    assert_equal(n_left, 1830)
    assert_equal(n_right, 513)

    projected = plane.project_on_normal(points)
    binning = [-160., -150., -140., -130., -120., -110., -100.,  -90.,  -80.,
     -70.,  -60.,  -50.,  -40.,  -30.,  -20.,  -10.,    0.,   10.]
    val, _ = np.histogram(projected, bins=binning)
    assert_array_equal(val,
                       [13,   65,  129,  247,  347,  581,  848, 1285, 1624, 2084, 1557,
                        1646, 1300, 1359, 1698, 1830,  513])


def test_success_function():
    rot_x, rot_y, rot_z = 4, 45, -21
    transl_x, transl_y, transl_z = 0, 0, 61
    bin_width = 10
    params = rot_x, rot_y, rot_z, transl_x, transl_y, transl_z
    res = _success_function(params, _get_points(), bin_width=bin_width)
    assert_equal(res, 513-1830)


def test_minimize():
    params = rot_x, rot_y, rot_z, transl_x, transl_y, transl_z = 4, 45, -21, 0, 0, 61

    result = _minimize(params, _get_points(), bin_width=10)
    assert_array_almost_equal(result,
                              [ 4.11038409e+00,  4.65181441e+01, -2.05934568e+01, -2.44713344e-04,
                                -2.71528635e-04,  6.88986409e+01])

def test__compute_probabilities():
    plane = test_module.CutPlane((1, 0, 0, 4), None, None, None)
    plane.histogram = MagicMock(return_value=[np.array([10, 2, 2])])
    plane._compute_probabilities()
    assert_equal(plane.minus_log_prob, 10.0)
    assert_equal(plane.status,
                 'The probability that there is in fact NO cut plane is high: -log(p) = 10.0 !')

    plane.histogram = MagicMock(return_value=[np.array([])])
    plane._compute_probabilities()
    ok_(np.isnan(plane.minus_log_prob))
    assert_equal(plane.status, 'The proba is NaN, something went wrong')


def test_plane_equation():
    # test invalid plane
    assert_raises(ValueError, test_module.PlaneEquation, 0, 0, 0, 4)

    equation = test_module.PlaneEquation.from_rotations_translations([0, 0, 2, 3, 4, 5])
    assert_array_equal(equation.coefs, [0, 0, 100, -500])
    assert_equal(str(equation), '(0.0) * X + (0.0) * Y + (100.0) * Z + (-500.0) = 0')

    assert_array_equal(equation.distance([[2,3,4], [2,3,7]]),
                       [1, 2])

    equation = test_module.PlaneEquation.from_rotations_translations([45, 0, 0, 3, 4, 5])
    assert_equal(equation.coefs[0], 0)
    assert_almost_equal(equation.coefs[1], -equation.coefs[2])
    assert_almost_equal(equation.coefs[3], -70.7106781186547)

    equation = test_module.PlaneEquation.from_rotations_translations([0, 0, 0, 0, 0, 0])
    assert_equal(equation.coefs[0], 0)
    assert_equal(equation.coefs[1], 0)
    assert_not_equal(equation.coefs[2], 0)
    assert_equal(equation.coefs[3], 0)


def test_from_json():
    filename = DATA / 'plane.json'

    def _assert_expected_plane(plane):
        assert_equal(plane.bin_width, 8)
        assert_array_equal(plane.coefs, [4, 7, 8, 1])
        assert_array_equal(plane.cut_leaves_coordinates,
                           [[10, 10, 10], [1, 0, 20]])
        assert_equal(plane.status, 'ok')
        assert_equal(plane.minus_log_prob, 70)

    _assert_expected_plane(
        test_module.CutPlane.from_json(str(filename))
    )

    _assert_expected_plane(
        test_module.CutPlane.from_json(filename)
    )

    with filename.open() as file_:
        dictionnary = json.load(file_)
        _assert_expected_plane(
            test_module.CutPlane.from_json(dictionnary)
        )


def test_find():
    filename = DATA / 'rotated.h5'
    neuron = nm.load_neuron(filename)
    result = test_module.CutPlane.find(neuron, bin_width=10).to_json()
    assert_equal(set(result.keys()),
                 {'details', 'cut-plane', 'cut-leaves', 'status'})

    assert_equal(result['status'], 'ok')
